#include <iostream>
#include <sstream>
#include <fstream>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_sf_gamma.h>
#include <float.h>
#include <math.h>

#include "beetle.H"

using namespace std;

// print software usage
void usage(char * name){
  fprintf(stdout,"\n%s version %s\n\n", name, VERSION);
  fprintf(stdout, "Usage: wfabc-simple -i infile [options]\n");
  fprintf(stdout, "-i     Infile with allele frequency data\n");
  fprintf(stdout, "-l     Number of SNP to analyze [0]\n");
  fprintf(stdout, "-n     Number of ABC simulations [1000]\n");
  fprintf(stdout, "-v     SD for slab mixture prior on s [0.1]\n");
  fprintf(stdout, "-h     estimate heterozygote effect [0]\n");  
  exit(1);
}

// ------ Functions for input and output ---------------

// read Ne and SNP allele frequency
void getdata(string file, dataset * data){
  int i, j;
  double x;
  string line, element;
  ifstream infile;
  istringstream stream;

  // open infile
  infile.open(file.c_str());
  if (!infile){
    cerr << "Cannot open file " << file << endl;
    exit(1);
  }

  // read Ne estimate, just L11
  getline(infile, line); // Ne information
  stream.str(line);
  for(j=0; j<NNE; j++){
    stream >> element;
    x = atof(element.c_str());
    //cerr << "ne " << j << ", " << x << endl;
    gsl_vector_set(data->varNe, j, x);
  }
  stream.clear();
    
  // read number of generations
  getline(infile, line); // Ne information
  stream.str(line);
  for(j=0; j<NNE; j++){
    stream >> element;
    x = atof(element.c_str());
    //cerr << "gen " << j << ", " << x << endl;
    gsl_vector_set(data->t, j, x);
  }
  stream.clear();


  // read and store the initial allele frequency for for SNP l
  for(i=0; i<data->snp; i++){
    getline(infile, line); // working through file until focal SNP
  }
  stream.str(line);
  stream.clear();
  // allele frequency
  stream >> element; // locus id
  stream >> element;
  data->p0 = atof(element.c_str());
  //cerr << "p0 = " << data->p0 << endl;
  infile.close();
}

// wrapper function for WF ABC simulations
void wfsim(dataset * data, parameters * parm){
  double p;
  
  // sample parameters from prior
  sampleparms(data, parm);
  // simulate model
  p = sim(data, parm) ;
  // print summary statistics
  cout << parm->mod << " " << parm->s << " " << parm->h << " " << p << endl;    
}

// sample model (constrained vs. unconstrained) and selection coefficients from priors
void sampleparms(dataset * data, parameters * parm){

  double u, s, h;
  double spike = 0.00714; // s dominates when >> 1/2N, so SD = 1/(2*70) or about 1/(2*Ne)
 			  // using same spike as for L14 
  // constant selection model
  parm->mod = 0; // constrain s to be same for all lines
  u = gsl_ran_flat(r, 0, 1); // spike or slab
  if (u < 0.5)
    s = gsl_ran_gaussian(r, parm->sigma); // 0-centered normal with slab SD
  else
    s = gsl_ran_gaussian(r, spike); // 0-centered normal with spike SD
  if (data->esth == 1){
    h = gsl_ran_flat(r, 0, 1); // uniform prior from 0 to 1
  }
  else
    h = 0.5;
  parm->s = s;
  parm->h = h;

}

// simulate evolution forward in time according to WF model
double sim(dataset * data, parameters * parm){

  int i;
  double p, pexp;
  double Ne;
  
  // just grap parental allele freq, no need for backward simulation
  p = data->p0;

  // simulate forward in time from ancestor to sample
  Ne = gsl_vector_get(data->varNe, 0);
  for(i=0; i<gsl_vector_get(data->t, 0); i++){
    pexp = calcDp(p, parm->s, parm->h);
    p = gsl_ran_binomial(r, pexp, (uint) (2 * Ne))/(2.0 * Ne);
  }
  return p;
}

 
// calculated expected change in allele freq. by selection
double calcDp(double p, double s, double h){
  double X, Y2, Z;
  double wbar, w[3];
  double dp;

  X = pow(p, 2);
  Y2 = 2 * p * (1-p);
  Z = pow(1-p,2);

  // w11 = fitness X = 1+s, w12 = fitness 2Y = 1+sh, w22 = fitness Z = 1
  w[0] = 1 + s; 
  w[1] = 1 + s * h;
  w[2] = 1;
  wbar = X * w[0] + Y2 * w[1] + Z * w[2];
  dp = (w[0] * X + w[1] * 0.5 * Y2 - p * wbar)/wbar;
  return (p + dp);
}
